{% load i18n %}

<!-- Skill Sets -->
<div role="tabpanel" class="tab-pane" id="skill_sets">
    <div class="table-responsive">
        <table class="table table-striped table-width-fix" id="tab_skill_sets">
            <thead>
                <tr>
                    <th>{% translate 'Skill Set Group' %}</th>
                    <th>{% translate 'Main' %}</th>
                    <th>{% translate 'State' %}</th>
                    <th>{% translate 'Organization' %}</th>
                    <th>{% translate 'Character' %}</th>
                    <th>{% translate 'Required Skills?' %}</th>
                    <th>{% translate 'Is Main?' %}</th>
                </tr>
            </thead>

            <tbody></tbody>
        </table>
    </div>
</div>
