# Generated by Django 4.0.10 on 2023-11-15 16:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("memberaudit", "0014_fix_default_permissions"),
    ]

    operations = [
        migrations.AddField(
            model_name="charactership",
            name="item_id",
            field=models.PositiveBigIntegerField(default=0),
            preserve_default=False,
        ),
    ]
